# Copyright (C) 2016 Xavier Gendre <xavier.gendre@math.univ-toulouse.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#####################################################
# Multiple Time Series Principal Component Analysis #
#####################################################

pkg_name <- "mtspca"

# Create a DataSet object from matrix data and data frame meta.
# The parameter n indicates the lengths of the successive time series. If n is
# a single integer, data must have a length equal to a multiple of n.
DataSet <- function(data, meta, n) {
  if (is.vector(data)) {
    data <- matrix(data, nrow=1L)
  } else {
    data <- as.matrix(data)
  }
  if (!is.double(data) && !is.integer(data)) {
    message(paste0(pkg_name, ": data must be double matrix"))
    return(NULL)
  }

  if (!is.data.frame(meta)) {
    message(paste0(pkg_name, ": meta must be data frame"))
    return(NULL)
  }

  n <- as.integer(n)
  if (!is.integer(n)   ||
      !is.vector(n)    ||
      !(length(n) > 0L) ||
      any(n <= 0L)) {
    message(paste0(pkg_name, ": n must be positive integer"))
    return(NULL)
  }

  if (nrow(data) != nrow(meta)) {
    message(paste0(pkg_name, ": data and meta sizes do not match"))
    return(NULL)
  }

  if (length(n) == 1L) {
    if (ncol(data) %% n) {
      message(paste0(pkg_name, ": invalid number of columns in data"))
      return(NULL)
    } else {
      n <- rep(n, ncol(data) %/% n)
    }
  } else {
    if (ncol(data) != sum(n)) {
      message(paste0(pkg_name, ": invalid number of columns in data"))
      return(NULL)
    }
  }

  return(structure(list(data=data, meta=meta, n=n), class="DataSet"))
}

# Return TRUE if obj is a DataSet, FALSE otherwise.
is.DataSet <- function(obj) {
  return(inherits(obj, "DataSet"))
}

# Override of length for DataSet
c.DataSet <- function(...) {
  args <- list(...)
  if (length(args) == 0L) {
    return(NULL)
  }

  ds_out <- NULL
  for (arg in args) {
    if (!is.DataSet(arg)) {
      message(paste0(pkg_name, ": invalid argument"))
      return(NULL)
    }

    if (is.null(ds_out)) {
      ds_out <- DataSet(data=arg$data, meta=arg$meta, n=arg$n)
    } else {
      if (any(arg$n != ds_out$n)) {
        message(paste0(pkg_name, ": bad number of columns"))
        return(NULL)
      }
      ds_out <- DataSet(data=rbind(ds_out$data, arg$data),
                        meta=rbind(ds_out$meta, arg$meta),
                        n=ds_out$n)
    }
  }

  return(ds_out)
}

# Override of length for DataSet
length.DataSet <- function(ds) {
  if (is.null(ds$data)) {
    return(0)
  }

  return(nrow(ds$data))
}

# Override of print for DataSet
print.DataSet <- function(ds) {
  cat("$data\n"); print(ds$data); cat("\n")
  cat("$meta\n"); print(ds$meta); cat("\n")
  cat("$n\n"); print(ds$n)
}

# Override of [] getter operator for DataSet
`[.DataSet` <- function(ds, i=NULL) {
  return(DataSet(data=ds$data[i,], meta=ds$meta[i,], n=ds$n))
}

# Override of [] setter operator for DataSet
`[<-.DataSet` <- function(ds, i=NULL, value=NULL) {
  # Not implemented
  message(paste0(pkg_name, ": writing in DataSet not implemented"))
  return(ds)
}

# Compute multiple time series PCA.
# Any serie containing at least one NA is omitted in the analysis.
# Parameters:
#   scale if TRUE, each column of curves is scaled in order to have a mean L2
#         norm equal to one (default TRUE)
# Output object:
#   $data     data matrix without NA
#   $meta     metadata associated to data without NA
#   $omitted  indices of removed time series due to NA
#   $means    matrix of means of each time serie
#   $n        lengths of each time series
#   $sd       vector of initial means of normalized L2 norms for each colum of
#             curves (does not exist if scale is FALSE)
#   $values   eigen values of PCA
#   $vectors  eigen vectors of PCA
#   $x        data in PCA space
mtspca <- function(ds, scale=TRUE) {
  if (!is.DataSet(ds)) {
    message(paste0(pkg_name, ": parameter ds has to be DataSet"))
    return(NULL)
  }

  # Returned object
  output <- structure(list(), class="mtspca")
  output$n <- ds$n

  # Remove NAs
  x <- na.omit(ds$data)
  output$omitted <- attr(x, "na.action")[]
  attr(x, "na.action") <- NULL

  # Cleaned data
  output$data <- x
  output$meta <- ds$meta[-output$omitted,]

  # Compute means of curves and center them
  output$means <- matrix(0, nrow=nrow(x), ncol=length(ds$n))
  for (i in seq_len(nrow(output$means))) {
    for (j in seq_len(ncol(output$means))) {
      start <- 1 + ifelse(j > 1, sum(ds$n[1:(j-1)]), 0)
      end <- sum(ds$n[1:j])
      output$means[i,j] <- mean(x[i, start:end])
      x[i, start:end] <- x[i, start:end] - output$means[i,j]
    }
  }

  # Scaling
  if (scale) {
    output$sd <- rep(0, length(ds$n))
    for (i in seq_along(output$sd)) {
      start <- 1 + ifelse(i > 1, sum(ds$n[1:(i-1)]), 0)
      end <- sum(ds$n[1:i])
      output$sd[i] <- sqrt(sum(x[,start:end]^2) / nrow(x))
      x[,start:end] <- x[,start:end] / output$sd[i]
    }
  }

  # Average curves
  output$centers <- colMeans(x)

  # PCA procedure
  x <- scale(x, center=TRUE, scale=FALSE)
  S <- t(x) %*% x / nrow(x)
  pca <- eigen(S)
  output$values <- pca$values
  output$vectors <- pca$vectors
  output$x <- x %*% pca$vectors

  # Due to approximations, some eigen values can be negative. We avoid such a
  # thing by thresholding these variances.
  output$values[output$values < 0] <- 0

  return(output)
}

# Return TRUE if obj is a mtspca, FALSE otherwise.
is.mtspca <- function(obj) {
  return(inherits(obj, "mtspca"))
}

# Override of length for mtspca
length.mtspca <- function(obj) {
  return(nrow(obj$x))
}

# Plot the data in the principal plane.
plot.mtspca <- function(obj, key=NULL, value=NULL, splitkey=NULL, splitkeyIsNum=TRUE, splitsize=1, q=0.75, visible=FALSE, col="black", lwd=1, lty=1, xlab=NULL, ylab=NULL, main=NULL, cex.lab=1) {
  plot(obj$x[,1:2],
       xlab=ifelse(is.null(xlab), "Component 1", xlab),
       ylab=ifelse(is.null(ylab), "Component 2", ylab),
       main=ifelse(is.null(main), "Principal plane", main),
       pch='.', col=ifelse(visible, "grey", "transparent"), cex.lab=cex.lab)

  group <- list()
  if (is.null(key)) {
    group[[1]] <- 1:length(obj)
  } else {
    if (is.null(value)) {
      group <- split(1:length(obj), obj$meta[key])
    } else {
      group[[1]] <- which(obj$meta[key] == value)
    }
  }

  for (i in 1:length(group)) {
    if (is.null(splitkey)) {
      suppressWarnings(mtspca.levelset(obj$x[group[[i]], 1:2], q, col=col[1+(i%%length(col))], lwd=lwd[1+(i%%length(lwd))], lty=lty[1+(i%%length(lty))]))
    } else {
      if (splitkeyIsNum) {
        bound <- quantile(unlist(obj$meta[splitkey])[group[[i]]], probs=seq(0, 1, length.out=splitsize+1))
        for (j in 1:splitsize) {
          filter <- bound[j] < obj$meta[splitkey] & obj$meta[splitkey] <= bound[j+1]
          if (j == 1) filter[obj$meta[splitkey] == bound[1]] <- TRUE
          suppressWarnings(mtspca.levelset(obj$x[intersect(group[[i]], which(filter)), 1:2], q, col=col[1+(j%%length(col))], lwd=lwd[1+(j%%length(lwd))], lty=lty[1+(j%%length(lty))]))
        }
      } else {
        subgroup <- split(1:length(obj), obj$meta[splitkey])
        for (j in 1:length(subgroup)) {
          suppressWarnings(mtspca.levelset(obj$x[intersect(group[[i]], subgroup[[j]]), 1:2], q, col=col[1+(j%%length(col))], lwd=lwd[1+(j%%length(lwd))], lty=lty[1+(j%%length(lty))]))
        }
      }
    }
  }
}

# Compute the coefficient of multiple correlation between the principal
# component of index pc and the set of variables associated to the i-th
# curve.
mtspca.cmc <- function(obj, pc, i) {
  if (!is.mtspca(obj)) {
    message(paste0(pkg_name, ": parameter obj has to be mtspca"))
    return(NA)
  }

  # Extract variables for the i-th curve
  start <- 1 + ifelse(i > 1, sum(obj$n[1:(i-1)]), 0)
  end <- sum(obj$n[1:i])
  X <- obj$data[,start:end]

  # Usual correlations
  ccor <- rep(0, ncol(X))
  for (k in seq_along(ccor)) ccor[k] <- cor(X[,k], obj$x[,pc])

  return(sqrt(t(ccor) %*% solve(cor(X)) %*% ccor))
}

# Graphics function to plot a q-level set of points given by the two first columns of x
library(MASS) # Temporary for kernel estimation
mtspca.levelset <- function(x, q, col="black", lty=1, lwd=1) {
  z <- kde2d(x[,1], x[,2], n=100) # Only one use of MASS function

  pp <- array()
  for (i in 1:nrow(x)) {
    z_x <- max(which(z$x < x[i,1]))
    z_y <- max(which(z$y < x[i,2]))
    pp[i] <- z$z[z_x, z_y]
  }
  cb <- quantile(pp, 1-q, na.rm=TRUE)

  contour(z, drawlabels=FALSE, levels=cb, col=col, add=TRUE, lty=lty, lwd=lwd)
}

# Compute the reading grid based on quantiles q
mtspca.readIngrid <- function(obj, q=c(0.1, 0.25, 0.5, 0.75, 0.9), n=10, limit=0.1) {
  if (!is.mtspca(obj)) {
    message(paste0(pkg_name, ": parameter obj has to be mtspca"))
    return(NULL)
  }

  qx <- quantile(obj$x[,1], probs=q)
  qy <- quantile(obj$x[,2], probs=q)

  inertia <- matrix(0, ncol=length(qx), nrow=length(qy))
  neigh <- array(0, c(length(qy), length(qx), n))

  #plot(obj$x[,1:2], xlab="Component 1", ylab="Component 2",
  #     main="Principal plane",
  #     pch='.', col="grey")
  for (i in seq_along(qx)) {
    #abline(v=qx[i], lty=3)
    for (j in seq_along(qy)) {
      #abline(h=qy[j], lty=3)

      dist2 <- (obj$x[,1] - qx[i])^2 + (obj$x[,2] - qy[j])^2
      inertia[i,j] <- sum(sort(dist2)[1:n]) / sum(obj$values[1:2])
      neigh[i,j,] <- order(dist2)[1:n]
      #for (k in neigh[i,j,]) {
      #  lines(c(qx[i], obj$x[k,1]), c(qy[j], obj$x[k,2]),
      #        col=ifelse(inertia[i,j]>limit, "blue", "red"))
      #}

      #text(qx[i], qy[j], labels=paste0(i, "/", j))
    }
  }

  limT <- range(obj$data[neigh,1:24] - matrix(rep(obj$means[neigh,1], 24), ncol=24))
  limRad <- range(obj$data[neigh,25:48] - matrix(rep(obj$means[neigh,2], 24), ncol=24))

  layout.matrix <- matrix(1:(length(qx)*length(qy)), ncol=length(qx))
  first_in_line <- apply(inertia <= limit, 2, function(v) { min(which(v)) })
  last_in_line <- apply(inertia <= limit, 2, function(v) { max(which(v)) })
  for (j in 1:ncol(layout.matrix)) layout.matrix[,j] <- rev(layout.matrix[,j])
  layout(
    rbind(layout.matrix, rep(0, length(qx))),
    respect=TRUE, heights=c(rep(1, length(qx)), 0.25))
  old_par <- par(mar=rep(0, 4))
  for (i in seq_along(qx)) {
    is_first_in_column = TRUE
    for (j in seq_along(qy)) {
      if (inertia[i,j]>limit) {
        plot(0, col="transparent", xlab="", ylab="", bty="n", xaxt="n", yaxt="n")
      } else {
        # Curves to plot
        vT <- rep(0, 24)
        for (k in 1:n) {
          vT <- vT + obj$data[neigh[i,j,k], 1:24] - obj$means[neigh[i,j,k], 1]
        }
        vT <- vT / n
        vRad <- rep(0, 24)
        for (k in 1:n) {
          vRad <- vRad + obj$data[neigh[i,j,k], 25:48] - obj$means[neigh[i,j,k], 2]
        }
        vRad <- vRad / n

        plot(vT, type="l", xlab="", ylab="", xaxt=ifelse(is_first_in_column, "s", "n"), yaxt="n", lwd=2, ylim=limT)
        is_first_in_column <- FALSE
        if (i == first_in_line[j]) {
          axis(side=2, at=c(-5,0,5), labels=c(-5,0,5))
        }
        par(new=TRUE)
        plot(vRad, type="l", xlab="", ylab="", xaxt="n", yaxt="n", lwd=2, lty=2, ylim=limRad)
        if (i == last_in_line[j]) {
          axis(side=4)
        }
      }
    }
  }
  par(old_par)
  layout(1)

  return(list(qx=qx, qy=qy, neigh=neigh, inertia=inertia))
}
