# Copyright (C) 2016 Xavier Gendre <xavier.gendre@math.univ-toulouse.fr>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

dir.create("output")
source("src/mtspca.R")

### Useful functions
####################

get_season <- function(month) {
  if (month %in% c(6, 7, 8)) { return("Summer") }
  if (month %in% c(1, 2, 12)) { return("Winter") }
  return("Intermediate")
}

### Data wrangling
##################

### Openfield data
raw_data <- read.table("data/Openfield_HEURE.dat", dec=",", head=T)
raw_data <- raw_data[raw_data$Hmeasure == 200,]
if (nrow(raw_data) %% 24) message("Bad line number")
n_day <- nrow(raw_data) %/% 24

# Initialization of data matrix and metadata
data_x <- matrix(c(raw_data[1:24, "T_openfield"], raw_data[1:24, "Rad_openfield"]), nrow=1)
data_meta <- data.frame(
  cover="openfield",
  site=raw_data[1,"Site"],
  year=raw_data[1,"Year"],
  month=raw_data[1,"Month"],
  day=raw_data[1,"Day"],
  season=get_season(raw_data[1,"Month"]),
  season_cover=paste0(get_season(raw_data[1,"Month"]), "_", "openfield"),
  part=1)

# Get the data
for (i in 2:n_day) {
  data_T <- raw_data[(1+24*(i-1)):(24*i), "T_openfield"]
  data_Rad <- raw_data[(1+24*(i-1)):(24*i), "Rad_openfield"]
  data_x <- rbind(data_x, c(data_T, data_Rad))

  meta <- data.frame(
    cover="openfield",
    site=raw_data[1+24*(i-1),"Site"],
    year=raw_data[1+24*(i-1),"Year"],
    month=raw_data[1+24*(i-1),"Month"],
    day=raw_data[1+24*(i-1),"Day"],
    season=get_season(raw_data[1+24*(i-1),"Month"]),
    season_cover=paste0(get_season(raw_data[1+24*(i-1),"Month"]), "_", "openfield"),
    part=1)
  data_meta <- rbind(data_meta, meta)
}

# Add data into a DataSet object
data_ds <- DataSet(data=data_x, meta=data_meta, n=24)

### Understorey data
raw_data <- read.table("data/CMP_NoVeg_HEURE.dat", dec=",", head=T)
raw_data <- raw_data[raw_data$Hmeasure == 200,]
if (nrow(raw_data) %% 24) message("Bad line number")
n_day <- nrow(raw_data) %/% 24

# Initialization of data matrix and metadata
data_x <- matrix(c(raw_data[1:24, "T_understorey"], raw_data[1:24, "Rad_understorey"]), nrow=1)
data_meta <- data.frame(
  cover="understorey",
  site=raw_data[1,"Site"],
  year=raw_data[1,"Year"],
  month=raw_data[1,"Month"],
  day=raw_data[1,"Day"],
  season=get_season(raw_data[1,"Month"]),
  season_cover=paste0(get_season(raw_data[1,"Month"]), "_", "understorey"),
  part=raw_data[1,"PARt_understorey"])

# Get the data
for (i in 2:n_day) {
  data_T <- raw_data[(1+24*(i-1)):(24*i), "T_understorey"]
  data_Rad <- raw_data[(1+24*(i-1)):(24*i), "Rad_understorey"]
  data_x <- rbind(data_x, c(data_T, data_Rad))

  meta <- data.frame(
    cover="understorey",
    site=raw_data[1+24*(i-1),"Site"],
    year=raw_data[1+24*(i-1),"Year"],
    month=raw_data[1+24*(i-1),"Month"],
    day=raw_data[1+24*(i-1),"Day"],
    season=get_season(raw_data[1+24*(i-1),"Month"]),
    season_cover=paste0(get_season(raw_data[1+24*(i-1),"Month"]), "_", "understorey"),
    part=raw_data[1+24*(i-1),"PARt_understorey"])
  data_meta <- rbind(data_meta, meta)
}

# Append data in the DataSet object
data_ds <- append(data_ds, DataSet(data=data_x, meta=data_meta, n=24))

# Variable names
colnames(data_ds$data) <- c(paste0("T", 1:24), paste0("Rad", 1:24))

### Data analysis
#################

result <- mtspca(data_ds)

# Point character and color lists
pch_list <- rep(0, nrow(result$data))
for (i in seq_along(pch_list)) {
  if (result$meta[i, "cover"] == "openfield") {
    pch_list[i] <- 4
  } else {
    if (result$meta[i, "cover"] == "understorey") {
      pch_list[i] <- 1
    } else {
      message("BUG")
    }
  }
}
col_list <- rep("black", nrow(result$data))
for (i in seq_along(col_list)) {
  if (result$meta[i, "season"] == "Winter") {
    col_list[i] <- grey(0)
  }
  if (result$meta[i, "season"] == "Summer") {
    col_list[i] <- grey(0.8)
  }
  if (result$meta[i, "season"] == "Intermediate") {
    col_list[i] <- grey(0.4)
  }
}

# PARt boxplots
pdf("output/Figure2.pdf", width=5, height=5)

boxplot(list(
    Winter=result$meta$part[result$meta$season_cover == "Winter_understorey"],
    Intermediate=result$meta$part[result$meta$season_cover == "Intermediate_understorey"],
    Summer=result$meta$part[result$meta$season_cover == "Summer_understorey"]
  ), ylab="PARt (%)", cex.lab=1.5)

dev.off()

# (T,Rad) means
pdf("output/Figure3.pdf",width=5,height=5)

par(mfrow=c(1,1),mar=par("mar")+c(0,1,0,0))
xlabel=expression(paste("Mean Radiation (J.", m^-2, ".", s^-1, ")"))
plot(result$means[,2], result$means[,1], pch=pch_list, col=col_list, xlab=xlabel, ylab="Mean Temperature (°C)", cex.lab=1.4)

dev.off()

# Plot a pair of (T,Rad) curves
plotTRad <- function(v) {
  # Size of vector v must be 48
  vT <- v[1:24]
  vRad <- v[25:48]
  old_par <- par(mar=c(4.5, 4.5, 0.5, 5))
  plot(vT, type="l", xlab="Time (h)", ylab="Temperature (°C)", las=1, lwd=2, cex.lab=1.3)
  par(new=TRUE)
  plot(vRad, type="l", xlab="", ylab="", xaxt="n", yaxt="n", lwd=2, lty=2)
  axis(side=4, las=1)
  ylabel=expression(paste("Radiation (J.", m^-2, ".", s^-1, ")"))
  mtext(ylabel, side=4, line=3, cex=1.3)
  par(old_par)
}

# Example of 8/12
pdf("output/Figure4.pdf", width=5,height=3)

plotTRad(result$data[result$meta$day == 12 & result$meta$month == 8 & result$meta$cover == "openfield",])

dev.off()

# Reading grid
pdf("output/Figure8tmp.pdf", width=6,height=5)
obj <- mtspca.readIngrid(result)
dev.off()

pdf("output/Figure5.pdf", width=7,height=5)

plot(result$x[,1:2], pch=pch_list, col=col_list, xlab="Component 1 (71%)", ylab="Component 2 (8%)", cex.lab=1.5)

for (i in seq_along(obj$qx)) {
  abline(v=obj$qx[i], lty=1, lwd=1.5)
  for (j in seq_along(obj$qy)) {
    abline(h=obj$qy[j], lty=1, lwd=1.5)
  }
}

for (i in seq_along(obj$qx)) {
  for (j in seq_along(obj$qy)) {
    points(obj$qx[i], obj$qy[j], cex=1.5, pch=16, col=ifelse(obj$inertia[i,j]>0.1, "red", "chartreuse3"))
  }
}

dev.off()

pdf("output/Figure6.pdf",width=7,height=5)
par(mfrow=c(1,1),mar=par("mar")+c(0,1,0,0))
plot(result,
     key="season_cover", value="Summer_understorey",
     splitkey="part", splitsize=4,
     col=c(grey(0.75), grey(0), grey(0.25), grey(0.5)),
     lwd=3, main=" ", cex.lab=1.5)
suppressWarnings(
  mtspca.levelset(result$x[result$meta$season_cover == "Summer_openfield", 1:2], 0.75, lty=2, lwd=3, col=grey(0.8))
)
for (i in seq_along(obj$qx)) {
  abline(v=obj$qx[i], lty=3)
  for (j in seq_along(obj$qy)) {
    abline(h=obj$qy[j], lty=3)
  }
}
dev.off()

# Openfield/Understorey level sets by season
pdf("output/Figure7.pdf",width=7,height=5)
par(mfrow=c(1,1),mar=par("mar")+c(0,1,0,0))
plot(result,
     splitkey="season_cover", splitkeyIsNum=FALSE,
     col=c("black", grey(0.4), grey(0.6), "black", grey(0.4), grey(0.6)),
     lty=c(1, 2, 2, 2, 1, 1),
     lwd=3, main=" ", cex.lab=1.5)
for (i in seq_along(obj$qx)) {
  abline(v=obj$qx[i], lty=3)
  for (j in seq_along(obj$qy)) {
    abline(h=obj$qy[j], lty=3)
  }
}
dev.off()
